//
//  OpenGL Demonstrator
//  Implementation of the Son Goku/Snowman interactive display demonstrator C++ classes
//
//  Created by J. M. Brisson Lopes
//  Updated on 2012/03/19
//  Updated on 2012/03/30
//  Last updated on 2013/05/04
//  Copyright (c) 2006-2013 J M. Brisson Lopes & Instituto Superior Tecnico. All rights reserved.
//


#include "goku.h"
#include "glm.h"

#include <cmath>                // Note C++ mat C library header style usage
#include <sstream>
#include <iomanip>
using namespace std;


// Displaylists manager (manages Wavefront OBJ, Snowman and tiled floor)


float CallLists::ModelSpeed = 0.05;

CallLists::CallLists()
{
    SnowManCallList = GokuCallList = FloorList= -1;
    ModelAngle      = 0.;
    drawSnowMan     = false;
    
    GLfloat matWhite[] = { 1., 1., 1., 1.};
    GLfloat matGrey[]  = { .7, .7, .7};
    GLfloat matBlack[] = { 0., 0., 0.,};
    GLfloat matNose[]  = { 1., .5, 0.};
    GLfloat matVerso[] = { 1., 1., 0., 1.};
    
    // Set up Goku from a Wavefront OBJ model using the GLM interface
    
    GokuCallList = glGenLists( 1);
	GLMmodel *goku = glmReadOBJ( (char*) "data/modelo_GT.obj");
    float sizes[3];
    glmDimensions( goku, sizes);
    glmScale( goku, 15. / sizes[1]);
	glmFacetNormals(goku);
    glmVertexNormals(goku, 90.0);
    glNewList( GokuCallList, GL_COMPILE);
    glPushMatrix();
    glTranslatef( 0., 7.1, 0.);
    glmDraw( goku, GLM_SMOOTH | GLM_MATERIAL);
    glPopMatrix();
    glEndList();
    glmDelete( goku);
    
    // Set up SnowMan ///////////////////////////////////////////////////////
    
    SnowManCallList = glGenLists( 1);
    glNewList( SnowManCallList, GL_COMPILE);
    
    glMaterialfv( GL_FRONT, GL_AMBIENT, matGrey);
    glMaterialfv( GL_FRONT, GL_DIFFUSE, matWhite);
    glMaterialfv( GL_FRONT, GL_SPECULAR, matGrey);
    glMaterialf(  GL_FRONT, GL_SHININESS, 5.);
    
    glPushMatrix();
    // Body
    glutSolidSphere(4.f,20,20);
    // Head
    glTranslatef(0.0f, 4.8f, 0.0f);
    glutSolidSphere(2.0f,40,40);
    // Eyes
    glPushMatrix();
    glMaterialfv( GL_FRONT, GL_DIFFUSE, matBlack);
    glTranslatef(0.48f, 0.80f, 1.6f);
    glutSolidSphere(0.4f,20,20);
    glTranslatef(-0.96f, 0.0f, 0.0f);
    glutSolidSphere(0.4f,20,20);
    glPopMatrix();
    // Nose
    glMaterialfv( GL_FRONT, GL_DIFFUSE, matNose);
    glTranslatef( 0.f, -0.1f, 0.85f);
    glRotatef(0.0f,1.0f, 0.0f, 0.0f);
    glutSolidCone(0.6f,2.5f,10,2);
    glPopMatrix();
    glEndList();
    
    
    // Set up the floor ////////////////////////////////
    
    FloorList = glGenLists( 1);
    glNewList( FloorList, GL_COMPILE);
    glMaterialf( GL_FRONT, GL_SHININESS, 1.5);
    int x, z;
    for( x=-8; x<8; x++)
    {
        for( z=-8; z<8; z++)
        {
            glPushMatrix();
            glTranslatef( x, -5., z);
            if ( ( (x+z) % 2) == 0 )
                glMaterialfv( GL_FRONT, GL_DIFFUSE, matBlack);
            else
                glMaterialfv( GL_FRONT, GL_DIFFUSE, matWhite);
            glBegin( GL_POLYGON);
            glNormal3f( 0., 1., 0.);
            glVertex3f( 0., 0., 0.);
            glVertex3f( 1., 0., 0.);
            glVertex3f( 1., 0., 1.);
            glVertex3f( 0., 0., 1.);
            glEnd();
            glPopMatrix();
        }
    }
    glMaterialfv( GL_FRONT, GL_DIFFUSE, matVerso);
    glPushMatrix();
    glTranslatef( 0., -5.2, 0.);
    glScalef( 1., .02, 1.);
    glutSolidCube( 16.);
    glPopMatrix();
    glEndList();
    
}


CallLists::~CallLists()
{
    if ( SnowManCallList != -1 )
        glDeleteLists( SnowManCallList, 1);
    if ( GokuCallList != -1 )
        glDeleteLists( GokuCallList, 1);
    if ( FloorList != -1 )
        glDeleteLists( FloorList, 1);
}


void CallLists::Draw( const int elapsedMilis)
{
    ModelAngle += (elapsedMilis * ModelSpeed);
    if ( ModelAngle >= 360. )
        ModelAngle -= 360.;
    unsigned int list = drawSnowMan ? SnowManCallList : GokuCallList;
    if ( list != -1)
    {
        glPushMatrix();
        glRotatef( ModelAngle, 0., 1., 0.);
        glCallList( list);
        glPopMatrix();
    }
    if ( FloorList != -1 )
        glCallList( FloorList);
}


const bool CallLists::SwapModel()
{
    drawSnowMan = !drawSnowMan;
    return drawSnowMan;
}


// Camera manager - defines and locates 4 different cameras


Camera::Camera()
{  mCameraAngle = 0.; SelectCamera( 1); }

int   Camera::minCameraType = 1, Camera::maxCameraType = 4;
float Camera::fov = 45.;
float Camera::cameraSpeed = .02;

const bool Camera::SelectCamera( const int cameraType)
{
    if ( cameraType >= minCameraType && cameraType <= maxCameraType ) {
        mCameraType = cameraType;
        SetProjection();
        return true;
    }
    return false;
}


const bool Camera::SetProjection() const
{
    float w = glutGet( GLUT_WINDOW_WIDTH),  h = glutGet( GLUT_WINDOW_HEIGHT);
    if ( w > 0. && h > 0. )
    {
        float ratio = w / h;
        glMatrixMode( GL_PROJECTION);
        glLoadIdentity();
        if ( mCameraType != 4 )
            gluPerspective( fov, ratio, 1., 100.);
        else
        {
            if ( w > h )
                glOrtho( -12. * ratio, 12. * ratio, -12., 12., -100., 100.);
            else
                glOrtho( -12., 12., -12. / ratio, 12. / ratio, -100., 100.);
        }
        return true;
    }
    return false;
}


const bool Camera::SetPosition( const int elapsedMilis)
{
    mCameraAngle += (cameraSpeed * elapsedMilis);
    if ( mCameraAngle >= 360. )
        mCameraAngle -= 360.;
    
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    switch ( mCameraType ) {
        case 1:
        case 2:
            gluLookAt( 20, 10, -20, 0, 0, 0, 0, 1, 0);
            break;
        case 3:
            glTranslatef( 0., 0., -28.);
            glRotatef( sin( mCameraAngle * 3.14159265 / 180. ) * 35. + 5., 1., 0., 0.);
            break;
        case 4:
            gluLookAt( 0., 20., 0., 0., 0., 0., 0., 0., 1.);
            break;
        default:
            break;
    }
    if ( mCameraType != 1 )
        glRotatef( mCameraAngle, 0., 1., 0.);
    return true;
}


// Lights manager - controls a spot lightsource, a direccional lightsource and the global ambient light


LightSource::LightSource()
{
    mMovingLightSource = true;
    mSpotLightSource   = false;
    mGlobalAmbientOn   = false;
    
    mLightSourceAngle   = 0.;
    GLfloat ambient[]   = { .2, .2, .2, 1.};
    GLfloat diffuse[]   = { 1., 1., 1., 1.};
    GLfloat specular[]  = { .8, .8, .8, 1.};
    
    glEnable( GL_LIGHTING);
    
    ToggleGlobalAmbient();
    
    glDisable( GL_LIGHT0);
    glLightfv( GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv( GL_LIGHT0, GL_SPECULAR, specular);
    glLightf(  GL_LIGHT0, GL_SPOT_CUTOFF, 45.);
    
    glDisable( GL_LIGHT1);
    glLightfv( GL_LIGHT1, GL_AMBIENT, ambient);
    glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse);
    glLightfv( GL_LIGHT1, GL_SPECULAR, specular);
    
    GLUquadric * quad = gluNewQuadric();
    gluQuadricDrawStyle( quad, GLU_FILL);
    gluQuadricNormals( quad, GLU_SMOOTH);
    float spotAmbient[]  = { .75,  .75,  .75,  1.};
    float spotDiffuse[]  = { .4,   .4,   .4,   1.};
    float spotSpecular[] = { .775, .775, .775, 1.};
    float noemission[]   = { 0., 0., 0., 1.};
    float emission[]     = { 1., 1., 1., 1.};
    
    // Display list to materialize a lamp
    
    SpotList = glGenLists( 1);
    glNewList( SpotList, GL_COMPILE);
    glMaterialfv( GL_FRONT, GL_AMBIENT,   spotAmbient);
    glMaterialfv( GL_FRONT, GL_DIFFUSE,   spotDiffuse);
    glMaterialfv( GL_FRONT, GL_SPECULAR,  spotSpecular);
    glMaterialf(  GL_FRONT, GL_SHININESS, 76.8);
    glMaterialfv( GL_FRONT, GL_EMISSION,  emission);
    glPushMatrix();
    glutSolidSphere( .5, 16, 16);
    glMaterialfv( GL_FRONT, GL_EMISSION, spotAmbient);
    glTranslatef( 0., .0, .1);
    gluDisk( quad, 0., 1.05, 64, 10);
    glPopMatrix();
    glPushMatrix();
    glMaterialfv( GL_FRONT, GL_EMISSION, noemission);
    gluCylinder( quad, 1.2, .32, .9, 64, 10);
    glTranslatef( 0., 0., .89);
    gluCylinder( quad, .32, .20, .325, 16, 10);
    glTranslatef( 0., 0., .32);
    gluDisk( quad, 0., .20, 16, 10);
    glPopMatrix();
    glEndList();
    
    gluDeleteQuadric( quad);
        
    ToggleLightSourceType();
}


const bool LightSource::ToggleLightSourceType()
{
    mSpotLightSource = !mSpotLightSource;
    if ( mSpotLightSource )
    {
        glEnable( GL_LIGHT0);
        glDisable( GL_LIGHT1);
    }
    else
    {
        glDisable( GL_LIGHT0);
        glEnable( GL_LIGHT1);
    }
    return mSpotLightSource;
}


const bool LightSource::ToggleLightSourceMovement()
{
    mMovingLightSource = !mMovingLightSource;
    return mMovingLightSource;
}


const bool LightSource::ToggleGlobalAmbient()
{
    mGlobalAmbientOn = !mGlobalAmbientOn;
    GLfloat * ambient = mGlobalAmbientOn ? globalAmbient : noGlobalAmbient;
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, ambient);
    return mGlobalAmbientOn;
}


const bool LightSource::PositionLightSource(const int elapsedMilis)
{
    if ( mMovingLightSource)
    {
        mLightSourceAngle += ( lightsourceSpeed * elapsedMilis );
        if ( mLightSourceAngle >= 360. )
            mLightSourceAngle -= 360.;
    }
    
    glPushMatrix();
    glRotatef( -mLightSourceAngle, 0, 1, 0);
    if ( mSpotLightSource )
    {
        lightsourcePosition[3] = 1.;
        glLightfv( GL_LIGHT0, GL_POSITION, lightsourcePosition);
        GLfloat direction[] = { -1, 0, 0};
        glLightfv( GL_LIGHT0, GL_SPOT_DIRECTION, direction);
    }
    else
    {
        lightsourcePosition[3] = 0.;
        glLightfv( GL_LIGHT1, GL_POSITION, lightsourcePosition);
    }
    
    glTranslatef( lightsourcePosition[0], lightsourcePosition[1], lightsourcePosition[2]);
    glRotatef( 90., 0., 1., 0.);
    glCallList( SpotList);
    glPopMatrix();
    
    return true;
}


GLfloat LightSource::lightsourceSpeed      = .05;
GLfloat LightSource::lightsourcePosition[] = { 16, 6, 0, 1};
GLfloat LightSource::globalAmbient[]       = { .3, .3, .3, 1.};
GLfloat LightSource::noGlobalAmbient[]     = { 0., 0., 0., 1.};


// OpenGLApplication

OpenGLApplication::OpenGLApplication()
{
    MyLists       = NULL;
    MyLightSource = NULL;
    MyCamera      = NULL;
    MyTimerOn     = true;
    FullScreenOn  = false;
    HUDon         = true;
    AxesOn        = false;
    lastTimeMilis           = 0;
    MyFrameCounter          = 0;
    LastFrameRateUpdateTime = 0;
}


void OpenGLApplication::Initialize( int argc, char **argv)
{
    glutInit( & argc, argv);
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    
    glutInitWindowSize( 600, 480);
    glutInitWindowPosition( 0, 0);
    glutCreateWindow( "Computer Graphics 2007/2013");
    
    MyCamera      = new Camera();
    MyLightSource = new LightSource();
    MyLists       = new CallLists();
    
    glutReshapeFunc( reshape);
    glutDisplayFunc( display);
    glutKeyboardFunc( keyboard);
    glutIdleFunc( NULL);
    glutTimerFunc( timerInterval, timer, 1);    // start in timer driven display mode

    glEnable( GL_DEPTH_TEST);
    glEnable( GL_NORMALIZE);
    glShadeModel( GL_SMOOTH);
    
    glutMainLoop();
}


OpenGLApplication::~OpenGLApplication()
{
    if ( MyCamera != NULL )
        delete MyCamera;
    if ( MyLightSource != NULL )
        delete MyLightSource;
    if ( MyLists != NULL )
        delete MyLists;
}


void OpenGLApplication::MyReshape(const int w, const int h)
{
    glViewport( 0, 0, w, h);
    MyCamera->SetProjection();
}


void OpenGLApplication::MyDisplay()
{
    // How long has it been since last time?
    int now       = glutGet( GLUT_ELAPSED_TIME);
    int elapsed   = now - lastTimeMilis;
    lastTimeMilis = now;
    
    // Compute frame rate
    MyFrameCounter++;
    if ( now >= (LastFrameRateUpdateTime + 1000) )
    {
        MyFrameRate = (1000. * MyFrameCounter) / ( now - LastFrameRateUpdateTime);
        LastFrameRateUpdateTime = now;
        MyFrameCounter = 0;
    }
    
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    MyCamera->SetPosition( elapsed);                     // Position camera
    MyLightSource->PositionLightSource( elapsed);        // Position lightsource
    MyLists->Draw( elapsed);                             // Draw objects
    if ( HUDon )
        HUD();                                           // Show status and commands
    
    glutSwapBuffers();
}


void OpenGLApplication::HUD()
{
    glDisable( GL_LIGHTING);
    if ( AxesOn ) {
        glLineWidth( 1.5);
        glColor3f( 1., 0., 0.);
        glBegin( GL_LINES);
        glVertex3f( 0., 0., 0.);
        glVertex3f( 20., 0., 0.);
        glEnd();
        glColor3f( 0., 1., 0.);
        glBegin( GL_LINES);
        glVertex3f( 0., 0., 0.);
        glVertex3f( 0., 20., 0.);
        glEnd();
        glColor3f( 1., 1., 0.);
        glBegin( GL_LINES);
        glVertex3f( 0., 0., 0.);
        glVertex3f( 0., 0., 20.);
        glEnd();
    }
    
    // Switch to raster mode (saving active projection)
    glMatrixMode( GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D( 0, glutGet( GLUT_WINDOW_WIDTH), 0, glutGet( GLUT_WINDOW_HEIGHT));
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    
    // Show commands
    glColor3f( 0., 1., 0.);
    DisplayString( 10, 18, commandLabel1);
    DisplayString( 10,  2, commandLabel2);
    
    // Show current settings and frame rate
    glRasterPos2i( 10, glutGet( GLUT_WINDOW_HEIGHT) - 20);
    ostringstream os;
    os.precision( 1);
    os << fixed << "Frame rate: " << fixed << setw(6) << MyFrameRate
       << " fps; Frame generation mode: " << (MyTimerOn ? "timer" : "idle");
    DisplayString( 5, glutGet( GLUT_WINDOW_HEIGHT)-20, os.str(), GLUT_BITMAP_HELVETICA_18);
    os.str( "");
    os << "Ambient: " << (MyLightSource->GlobalAmbientOn() ? "on" : "off")
       << "; Light source: " << (MyLightSource->SpotLightOn()? "spot" : "direccional")
    << "; Camera " << MyCamera->CameraNo();
    DisplayString( 5, glutGet( GLUT_WINDOW_HEIGHT)-38, os.str(), GLUT_BITMAP_HELVETICA_18);
    
    // Switch back to normal mode (restoring projection)
    glMatrixMode( GL_PROJECTION);
    glPopMatrix();
    glEnable( GL_LIGHTING);
}


void OpenGLApplication::DisplayString( const int x, const int y, const string &label, const void * font)
{
    void * fontToUse = (font != NULL) ? (void*) font : GLUT_BITMAP_HELVETICA_12;
    glRasterPos2i( x, y);
    for ( int i=0; i<label.size(); i++)
        glutBitmapCharacter( fontToUse, label[i]);
}


void OpenGLApplication::MyKeyboard( const unsigned char key, const int x, const int y)
{
    switch ( tolower( key) ) {
        case 27:              // To end program
            exit( 0);
            break;
        case 'f':             // Full screen / Normal screen
            FullScreenOn = !FullScreenOn;
            if ( FullScreenOn )
                glutFullScreen();
            else
            {
                glutReshapeWindow( glutGet( GLUT_INIT_WINDOW_WIDTH), glutGet( GLUT_INIT_WINDOW_HEIGHT));
                glutPositionWindow( glutGet( GLUT_INIT_WINDOW_X), glutGet( GLUT_INIT_WINDOW_Y));
            }
            break;
        case '1':             // Camera selection
        case '2':
        case '3':
        case '4':
            MyCamera->SelectCamera( key - '0');
            break;
        case 't':             // Timer mode / idle mode
            if ( MyTimerOn )
                glutIdleFunc( idle);
            else
            {
                glutIdleFunc( NULL);
                glutTimerFunc( timerInterval, timer, 0);
            }
            MyTimerOn = !MyTimerOn;
            break;
        case 's':              // Change model
            MyLists->SwapModel();
            break;
        case 'l':               // Change lightsource type (spot/direccional)
            MyLightSource->ToggleLightSourceType();
            break;
        case 'm':               // Move / stop lightsource
            MyLightSource->ToggleLightSourceMovement();
            break;
        case 'a':               // Turn global ambient light on/off
            MyLightSource->ToggleGlobalAmbient();
            break;
        case 'h':
            HUDon = !HUDon;
            break;
        case 'x':
            AxesOn = !AxesOn;
            break;
        default:
            break;
    }
}


void OpenGLApplication::MyIdle()
{   glutPostRedisplay();  }


void OpenGLApplication::MyTimer(const int value)
{
    if ( MyTimerOn )
    {
        glutTimerFunc( timerInterval, timer, 0);
        glutPostRedisplay();
    }
}

OpenGLApplication OpenGLApplication::MyApplication;
int               OpenGLApplication::timerInterval = 40;
string            OpenGLApplication::commandLabel1 = "Toggles: 1-4 cameras, a-ambient light, l-light type, m-moving light, s-model, t-timer/idle",
OpenGLApplication::commandLabel2 ="Esc to exit, f-normal/full screen, h-HUD on/off"; 


// main


int main(int argc,char** argv)
{   
    OpenGLApplication::MyApplication.Initialize( argc, argv);
	return 0;
}
